/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original developer)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLCSVParser.h"
#include <prdtoa.h>
#include <nsCRT.h>
#include <math.h>
#include <nsIServiceManager.h>
#include <nsMemory.h>

mozCellMLCSVParser::mozCellMLCSVParser()
  : Capacity(16), Size(0)
{
  Data = new double[16];
}

mozCellMLCSVParser::~mozCellMLCSVParser()
{
  delete [] Data;
}

NS_IMPL_ISUPPORTS1(mozCellMLCSVParser, mozCellMLICSVParser)

NS_IMETHODIMP
mozCellMLCSVParser::FetchFields(PRUint32* nvalues,
                                char*** titles)
{
  nsCOMPtr<nsIMemory> mm(do_GetService("@mozilla.org/xpcom/memory-service;1"));
  *nvalues = Names.Count();
  *titles = (char**)mm->Alloc(Names.Count() * sizeof(char*));

  PRInt32 i;
  for (i = 0; i < Names.Count(); i++)
    (*titles)[i] = nsCRT::strdup(Names[i]->get());

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLCSVParser::GetRowCount(PRUint32* count)
{
  *count = Size / Names.Count();
  return NS_OK;
}

void
mozCellMLCSVParser::EnsureCapacity()
{
  // Size is the old size, we want to grow...
  if (Capacity <= Size)
  {
    if (Capacity <= 1024)
      Capacity *= 2;
    else
      Capacity += (Capacity / 2);
    double* OldData = Data;
    Data = new double[Capacity];
    // XXX can we assume memcpy is available?
    memcpy(Data, OldData, sizeof(double) * Size);
    delete [] OldData;
  }
}

void
mozCellMLCSVParser::NameCompleted()
{
  Names.AppendCString(TmpString);
}

#ifndef NAN
#include <LIMITS>
#define NAN std::numeric_limits<double>::quiet_NaN()
#endif

void
mozCellMLCSVParser::ValueCompleted()
{
  if (nThisRow >= (PRUint32)Names.Count())
    return;
  EnsureCapacity();
  const char* tptr = TmpString.get();
  char* endptr;
  Data[Size++] = PR_strtod(tptr, &endptr);
  if (endptr == tptr)
    Data[Size - 1] = NAN;
  nThisRow++;
}

void
mozCellMLCSVParser::RowCompleted()
{
  PRUint32 deficit = Names.Count() - nThisRow;
  PRUint32 i;
  for (i = 0; i < deficit; i++)
  {
    EnsureCapacity();
    Data[Size++] = NAN;
  }
  nThisRow = 0;
}

PRBool
mozCellMLCSVParser::IsNameChar(char c)
{
  if (c >= 'A' && c <= 'Z')
    return PR_TRUE;
  if (c >= 'a' && c <= 'z')
    return PR_TRUE;
  if (c >= '0' && c <= '9')
    return PR_TRUE;
  if (c == '_')
    return PR_TRUE;
  return PR_FALSE;
}

PRBool
mozCellMLCSVParser::IsValueChar(char c)
{
  if (c == '.' || (c >= '0' && c <= '9'))
    return PR_TRUE;
  return PR_FALSE;
}

void
mozCellMLCSVParser::StartNameChar(char c)
{
  if (IsNameChar(c))
  {
    ParserState = UNQUOTED_NAME;
    TmpString = c;
    return;
  }
  else if (c == '\"')
  {
    ParserState = QUOTED_NAME;
    TmpString = "";
    return;
  }
  else if (c == '\r' || c == ' ' || c == '\t')
    return;
  else if (c == ',')
  {
    TmpString = "";
    NameCompleted();
  }
  else if (c == '\n')
    ParserState = START_VALUE;
  else
  {
#ifdef DEBUG_MODE
    printf("Unexpected character %c starting name.\n");
#endif
    ParserState = PARSE_FAILURE;
  }
}

void
mozCellMLCSVParser::QuotedNameChar(char c)
{
  if (IsNameChar(c))
  {
    TmpString += c;
    return;
  }
  else if (c == '\"')
  {
    ParserState = END_QUOTED_NAME;
    NameCompleted();
    return;
  }
  else
    ParserState = PARSE_FAILURE;
}

void
mozCellMLCSVParser::UnquotedNameChar(char c)
{
  if (IsNameChar(c))
  {
    TmpString += c;
    return;
  }
  else if (c == ',')
  {
    NameCompleted();
    ParserState = START_NAME;
    return;
  }
  else if (c == '\r' || c == ' ' || c == '\t')
  {
    NameCompleted();
    ParserState = END_QUOTED_NAME;
  }
  else if (c == '\n')
  {
    NameCompleted();
    ParserState = START_VALUE;
  }
  else
  {
    ParserState = PARSE_FAILURE;
  }
}

void
mozCellMLCSVParser::EndQuotedNameChar(char c)
{
  if (c == ',')
    ParserState = START_NAME;
  else if (c == '\r' || c == '\t' || c == ' ')
    return;
  else if (c == '\n')
    ParserState = START_VALUE;
  else
    ParserState = PARSE_FAILURE;
}

void
mozCellMLCSVParser::StartValueChar(char c)
{
  if (IsValueChar(c))
  {
    ParserState = UNQUOTED_VALUE;
    TmpString = c;
    return;
  }
  else if (c == '\"')
  {
    ParserState = QUOTED_VALUE;
    TmpString = "";
    return;
  }
  else if (c == '\r' || c == ' ' || c == '\t')
    return;
  else if (c == ',')
  {
    TmpString = "";
    ValueCompleted();
  }
  else if (c == '\n')
  {
    RowCompleted();
    ParserState = START_VALUE;
  }
  else
    ParserState = PARSE_FAILURE;
}

void
mozCellMLCSVParser::QuotedValueChar(char c)
{
  if (IsValueChar(c))
  {
    TmpString += c;
    return;
  }
  else if (c == '\"')
  {
    ParserState = END_QUOTED_VALUE;
    ValueCompleted();
    return;
  }
  else
    ParserState = PARSE_FAILURE;
}

void
mozCellMLCSVParser::UnquotedValueChar(char c)
{
  if (IsValueChar(c))
  {
    TmpString += c;
    return;
  }
  else if (c == ',')
  {
    ValueCompleted();
    ParserState = START_VALUE;
    return;
  }
  else if (c == '\r' || c == ' ' || c == '\t')
  {
    ValueCompleted();
    ParserState = END_QUOTED_VALUE;
  }
  else if (c == '\n')
  {
    ValueCompleted();
    RowCompleted();
    ParserState = START_VALUE;
  }
  else
  {
    ParserState = PARSE_FAILURE;
  }
}

void
mozCellMLCSVParser::EndQuotedValueChar(char c)
{
  if (c == ',')
    ParserState = START_VALUE;
  else if (c == '\r' || c == '\t' || c == ' ')
    return;
  else if (c == '\n')
  {
    RowCompleted();
    ParserState = START_VALUE;
  }
  else
    ParserState = PARSE_FAILURE;
}

void
mozCellMLCSVParser::ProcessCharacter(char c)
{
  switch (ParserState)
  {
  case START_NAME:
    StartNameChar(c);
    break;
  case QUOTED_NAME:
    QuotedNameChar(c);
    break;
  case UNQUOTED_NAME:
    UnquotedNameChar(c);
    break;
  case END_QUOTED_NAME:
    EndQuotedNameChar(c);
    break;
  case START_VALUE:
    StartValueChar(c);
    break;
  case QUOTED_VALUE:
    QuotedValueChar(c);
    break;
  case UNQUOTED_VALUE:
    UnquotedValueChar(c);
    break;
  case END_QUOTED_VALUE:
    EndQuotedValueChar(c);
    break;
  case PARSE_FAILURE:
    return;
  }
}

NS_IMETHODIMP
mozCellMLCSVParser::ParseCSV(nsIInputStream* is)
{
  char buf[1024];
  ParserState = START_NAME;
  nThisRow = 0;
  Size = 0;
  while (ParserState != PARSE_FAILURE)
  {
    PRUint32 len;
    is->Read(buf, 1024, &len);
    if (len == 0)
      break;
    PRUint32 i;
    for (i = 0; i < len; i++)
      ProcessCharacter(buf[i]);
  }
  if (ParserState == PARSE_FAILURE)
    return NS_ERROR_FAILURE;
  else if (ParserState == START_NAME)
    ;
  else if (ParserState == QUOTED_NAME || ParserState == QUOTED_VALUE)
  {
#ifdef DEBUG_MODE
    printf("End-of-file encountered during quoted name or value.\n");
#endif
    // This is a parse error...
    return NS_ERROR_FAILURE;
  }
  else if (ParserState == UNQUOTED_NAME)
  {
    NameCompleted();
  }
  else if (ParserState == START_VALUE)
  {
    if (nThisRow != 0)
      RowCompleted();
  }
  else if (ParserState == UNQUOTED_VALUE)
  {
    ValueCompleted();
    RowCompleted();
  }
  else if (ParserState == END_QUOTED_VALUE)
    RowCompleted();

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLCSVParser::FetchData(
                              PRUint32 row,
                              PRUint32 *nvalues,
                              double **data
                             )
{
  PRUint32 startSize = Names.Count() * row;
  PRUint32 endSize = startSize + Names.Count();
  if (endSize > Size)
  {
    *nvalues = 0;
    return NS_OK;
  }

  *nvalues = Names.Count();
  nsCOMPtr<nsIMemory> mm(do_GetService("@mozilla.org/xpcom/memory-service;1"));
  *data = (double*)mm->Alloc(Names.Count() * sizeof(double));

  PRInt32 i;
  for (i = 0; i < Names.Count(); i++)
    (*data)[i] = Data[startSize + i];
  return NS_OK;
}
